import logging

from py_fastapi_logging.utils.extra import get_extra_from_environ


class BaseFormatter(logging.Formatter):
    def format(self, record: logging.LogRecord, *args, **kwargs) -> str:
        record: logging.LogRecord = self._update_extra(record)
        log: str = self._format_log(record)
        return log

    def _update_extra(self, record: logging.LogRecord) -> logging.LogRecord:
        extra = get_extra_from_environ()
        for extra_key, extra_value in extra.items():
            if not hasattr(record, extra_key):
                setattr(record, extra_key, extra_value)
        return record



import datetime
import json
import logging
import traceback

from py_fastapi_logging.formatters.base import BaseFormatter
from py_fastapi_logging.schemas.base import BaseJsonLogSchema
from py_fastapi_logging.utils.extra import get_env_extra


class JSONLogFormatter(BaseFormatter):
    @staticmethod
    def _format_log(record: logging.LogRecord) -> dict:
        now = datetime.datetime.fromtimestamp(record.created).astimezone().replace(microsecond=0).isoformat()

        json_log_fields = BaseJsonLogSchema(
            thread=record.process,
            timestamp=now,
            level=record.levelno,
        )

        for key in get_env_extra().keys():
            if hasattr(record, key):
                json_log_fields[key] = getattr(record, key)
            elif key == "progname":
                json_log_fields[key] = record.module
        if record.exc_info:
            json_log_fields["exceptions"] = traceback.format_exception(*record.exc_info)

        elif record.exc_text:
            json_log_fields["exceptions"] = record.exc_text

        if hasattr(record, "payload"):
            json_log_fields["payload"] = record.payload
        elif hasattr(record, "message"):
            json_log_fields["payload"] = {"message": record.message}
        elif hasattr(record, "msg"):
            json_log_fields["payload"] = {"message": record.msg}
        return json.dumps(json_log_fields, ensure_ascii=False)



import functools
import logging
import os
import pathlib
from logging.handlers import RotatingFileHandler, TimedRotatingFileHandler

from py_fastapi_logging.formatters.json import JSONLogFormatter
from py_fastapi_logging.formatters.simple import SimpleLogFormatter
from py_fastapi_logging.utils.extra import set_extra_to_environ


class LogConfigure:
    DEFAULT_OUTPUTS = "json,stderr"

    FORMAT_MAP = {"console": "stderr"}

    def __init__(
        self,
        app_name=None,
        filename=None,
        level=None,
        rotate_mb=None,
        rotate_when=None,
        backup_count=7,
        json_multiline=False,
        request_id_support=True,
        log_output=None,
        colored=True,
        log_dir=None,
        cleanup=False,
        exclude_fields=None,
    ):
        self.app_name = app_name
        self._level = level
        self.rotate_mb = rotate_mb
        self.rotate_when = rotate_when
        self.backup_count = backup_count
        self.json_multiline = json_multiline
        self._log_output = log_output
        self._colored = colored
        self._log_dir = log_dir
        self._filename = pathlib.Path(filename) if filename else None
        self.request_id_support = request_id_support
        self._apply_suffix = False
        self.cleanup = cleanup
        self.exclude_fields = exclude_fields

    @property
    def base_app_name(self):
        return self.app_name.split(".", 1)[0]

    @functools.cached_property
    def directory(self):
        log_dir = os.environ.get("LOG_DIR")
        if log_dir:
            return pathlib.Path(log_dir)

        if self._filename and self._filename.is_absolute():
            directory = self._filename.parent
            return directory

        return pathlib.Path("/var/log/", self.base_app_name)

    def get_filename(self, suffix):
        if self._filename:
            filename = self._filename
        else:
            filename = os.environ.get("LOG_FILENAME", f"{self.base_app_name}.log")
        filename = self.directory / filename
        if self._apply_suffix:
            filename.with_suffix(suffix)

        self._apply_suffix = True
        return filename

    def get_exclude_fields(self):
        if self.exclude_fields:
            return self.exclude_fields
        else:
            return os.environ.get("LOG_EXCLUDE_FIELDS", "").split(",")

    def _get_file_handler(self, suffix):
        filename = self.get_filename(suffix)
        if self.rotate_mb:
            handler = RotatingFileHandler(
                filename,
                maxBytes=self.rotate_mb * 1024 * 1024,
                backupCount=self.backup_count,
                encoding="utf-8",
            )

        elif self.rotate_when:
            handler = TimedRotatingFileHandler(
                filename,
                when=self.rotate_when,
                backupCount=self.backup_count,
                encoding="utf-8",
            )
        else:
            handler = logging.FileHandler(filename, encoding="utf-8")
        return handler

    def get_handler(self, name):
        if name == "simple":
            handler = self._get_file_handler(".txt")
            formatter = SimpleLogFormatter()
            handler.setFormatter(formatter)
        elif name == "json-stdout":
            handler = logging.StreamHandler()
            formatter = JSONLogFormatter()
            handler.setFormatter(formatter)
        elif name == "json":
            handler = self._get_file_handler(".json")
            formatter = JSONLogFormatter()
            handler.setFormatter(formatter)
        elif name == "stderr":
            handler = logging.StreamHandler()
            formatter = SimpleLogFormatter()
            handler.setFormatter(formatter)
        else:
            raise ValueError(f"unknown handler type: '{name}'")

        return handler

    @property
    def level(self):
        return os.environ.get("LOG_LEVEL", self._level or "INFO").upper()

    def normalize_format_names(self, names: str):
        names = names.lower().split(",")
        return [self.FORMAT_MAP.get(name, name) for name in names]

    @property
    def formats(self):
        if self._log_output:
            names = self._log_output
        else:
            names = os.environ.get("LOG_OUTPUT") or os.environ.get("LOG_FORMAT")
        if not names:
            names = self.DEFAULT_OUTPUTS
        return self.normalize_format_names(names)

    def get_handlers(self, names):
        handlers = [self.get_handler(name) for name in names]
        return handlers

    @property
    def default_logger_level(self):
        level = os.environ.get("LOGGING_DEFAULT_LEVEL") or os.environ.get("LOG_DEFAULT_LEVEL")
        if level:
            level = level.upper()
        return level

    @property
    def default_logger_output(self):
        return os.environ.get("LOGGING_DEFAULT_HANDLER") or os.environ.get("LOG_DEFAULT_OUTPUT")


def _init_logger(config: LogConfigure):
    loggers = [logging.getLogger(name) for name in logging.root.manager.loggerDict]

    if config.cleanup:
        for logger in loggers:
            for h in logger.handlers[:]:
                logger.removeHandler(h)
                h.close()

    config.directory.mkdir(exist_ok=True)

    handlers = config.get_handlers(config.formats)

    logging.basicConfig(handlers=handlers, level=config.level)
    set_extra_to_environ("progname", config.app_name)


def init_logger(app_name: str):
    return _init_logger(LogConfigure(app_name=app_name, cleanup=True))



import datetime
import logging
import traceback

from py_fastapi_logging.formatters.base import BaseFormatter


class SimpleLogFormatter(BaseFormatter):
    @staticmethod
    def _format_log(record: logging.LogRecord) -> str:
        now = datetime.datetime.fromtimestamp(record.created).astimezone().replace(microsecond=0).isoformat()

        log_str = f"[{now}] "
        log_str += f"{record.levelname} "
        if hasattr(record, "progname"):
            progname = record.progname
        else:
            progname = record.module
        log_str += f" -- {progname}: "

        if hasattr(record, "request_id"):
            log_str += f"[{record.request_id}] "
        if hasattr(record, "tags"):
            log_str += f"{record.tags} "
        if hasattr(record, "message"):
            msg = record.message % record.args
            log_str += f"{msg}"

        elif hasattr(record, "msg"):
            msg = record.msg % record.args
            log_str += f"{msg}"

        if record.exc_info:
            log_str += f"\n{traceback.format_exception(*record.exc_info)}\n"

        elif record.exc_text:
            log_str += f"\n{record.exc_text}\n"

        if hasattr(record, "payload"):
            log_str += f"{record.payload}"

        return log_str



import http
import json
import logging
import math
import time

from fastapi import Request, Response
from starlette.middleware.base import BaseHTTPMiddleware, DispatchFunction, RequestResponseEndpoint
from starlette.types import ASGIApp

from py_fastapi_logging.config.config import LogConfigure, init_logger
from py_fastapi_logging.middlewares.utils.request_id import REQUEST_ID_HEADER, generate_request_id
from py_fastapi_logging.schemas.request import RequestPayload
from py_fastapi_logging.schemas.response import ResponsePayload
from py_fastapi_logging.utils.extra import set_extra_to_environ


class LoggingMiddleware(BaseHTTPMiddleware):
    _app_name: str = ""
    _prefix: str = ""
    _logger: logging.Logger = None

    def __init__(
        self,
        app: ASGIApp,
        app_name: str,
        dispatch: DispatchFunction = None,
        prefix: str = None,
        logger: logging.Logger = None,
    ):
        if not logger:
            init_logger(app_name=app_name)
            logger = logging.getLogger("default")
        self.app = app
        self._app_name = app_name
        self._prefix = prefix
        self._logger = logger
        self.dispatch_func = self.dispatch if dispatch is None else dispatch

    @staticmethod
    def get_request_id_header(headers):
        for name, value in headers.items():
            if name.lower() == REQUEST_ID_HEADER.lower():
                try:
                    return value
                except UnicodeDecodeError:
                    return ""

    @staticmethod
    async def get_protocol(request: Request) -> str:
        protocol = str(request.scope.get("type", ""))
        http_version = str(request.scope.get("http_version", ""))
        if protocol.lower() == "http" and http_version:
            return f"{protocol.upper()}/{http_version}"
        return ""

    @staticmethod
    async def set_body(request: Request, body: bytes) -> None:
        async def receive() -> dict:
            return {"type": "http.request", "body": body}

        request._receive = receive

    async def get_body(self, request: Request) -> bytes:
        body = await request.body()
        await self.set_body(request, body)
        return body

    async def dispatch(self, request: Request, call_next: RequestResponseEndpoint) -> Response:
        start_time = time.time()
        exception_object = None
        try:
            raw_request_body = await request.body()
            await self.set_body(request, raw_request_body)
            raw_request_body = await self.get_body(request)
            request_body = raw_request_body.decode()
        except Exception:
            request_body = ""
        request_headers: dict = dict(request.headers.items())
        request_id = self.get_request_id_header(request_headers) or generate_request_id(prefix=self._prefix)
        set_extra_to_environ("request_id", request_id)

        server: tuple = request.get("server", ("localhost", 80))
        request_params = request.query_params

        try:
            response = await call_next(request)
        except Exception as ex:
            response_body = bytes(http.HTTPStatus.INTERNAL_SERVER_ERROR.phrase.encode())
            response = Response(
                content=response_body,
                status_code=http.HTTPStatus.INTERNAL_SERVER_ERROR.real,
            )
            exception_object = ex
        else:
            response_body = b""
            async for chunk in response.body_iterator:
                response_body += chunk

            response = Response(
                content=response_body,
                status_code=response.status_code,
                headers=dict(response.headers),
                media_type=response.media_type,
            )
        duration: int = math.ceil((time.time() - start_time) * 1000)

        request_log_payload = RequestPayload(
            method=request.method,
            path=request.url.path,
            host=f"{server[0]}:{server[1]}",
            params=dict(request_params) if request_params is not None else {},
            body=request_body,
        )

        response_log_payload = ResponsePayload(
            status=response.status_code,
            response_time=f"{duration}ms",
            response_body=response_body.decode(),
        )

        logger_method = getattr(self._logger, "info" if not exception_object else "error")

        extra_fields = {
            "tags": ["API", "REQUEST"],
            "payload": self._filter_private_data(request_log_payload),
            "to_mask": True,
            "progname": self._app_name,
            "request_id": request_id,
        }

        logger_method(
            f"REQUEST {request.method} {request.url.path}",
            extra=extra_fields,
            exc_info=exception_object,
        )

        extra_fields["tags"] = ["API", "RESPONSE"]
        extra_fields["payload"] = self._filter_private_data(response_log_payload)

        logger_method(
            f"RESPONSE {response.status_code} {request.url.path}",
            extra=extra_fields,
            exc_info=exception_object,
        )

        return response

    def _filter_private_data(self, payload: dict | None):
        _filtered_fields = LogConfigure().get_exclude_fields()
        _marker_filtered = "[filtered]"
        if "params" in payload.keys() and payload["params"] is not None and _filtered_fields is not None:
            for key in list(payload["params"]):
                if key in _filtered_fields:
                    payload["params"][key] = _marker_filtered
            payload = self._filter_json_field("body", payload, _filtered_fields, _marker_filtered)
        elif "response_body" in payload.keys() and _filtered_fields is not None:
            payload = self._filter_json_field("response_body", payload, _filtered_fields, _marker_filtered)
        return payload

    def _filter_json_field(self, key, payload, _filtered_fields, _marker_filtered="[filtered]"):
        try:
            body_dict = json.loads(payload[key])
            if body_dict is not None:
                for key in list(body_dict):
                    if key in _filtered_fields:
                        body_dict[key] = _marker_filtered
                payload[key] = json.dumps(body_dict)
        except json.decoder.JSONDecodeError:
            pass
        return payload



from typing import Optional, TypedDict, Union


class BaseJsonLogSchema(TypedDict):
    thread: Union[int, str]
    level: int
    request_id: Optional[str]
    progname: Optional[str]
    timestamp: str
    exceptions: Union[list[str], str] = None



from uuid import uuid4

REQUEST_ID_HEADER = "X-Request-Id"
DEFAULT_DELIMITER = "#"


def generate_request_id(*, prefix=None, suffix=None, delimiter=DEFAULT_DELIMITER):
    prefix = f"{prefix}{delimiter}" if prefix else ""
    suffix = f"{delimiter}{suffix}" if suffix else ""
    request_id = "".join((prefix, str(uuid4()), suffix))
    return request_id



from setuptools import setup, find_packages

setup(
    name='py-fastapi-logging',
    version='0.1.0',
    packages=find_packages(exclude=['tests']),
    url='',
    license='',
    author='Dmitry Akhnazarov',
    author_email='wikedwolf@ya.ru',
    description='FastAPI logging requests'
)



from typing import Optional, TypedDict


class RequestPayload(TypedDict):
    method: str
    path: str
    host: str
    subject_type: Optional[str]
    subject_id: Optional[str]
    params: Optional[dict]
    body: str



import os


def get_env_extra():
    res = {
        "request_id": "LOGTMP-X-REQUEST-ID",
        "progname": "LOGTMP-PROGNAME",
    }
    env_extra_str = os.environ.get("LOG_ENV_EXTRA")
    if env_extra_str:
        extra_list = env_extra_str.split(",")
        for extra in extra_list:
            key, val = extra.split(":")
            res[key.strip()] = val.strip()
    return res


def get_extra_from_environ():
    res = {}
    for key, ename in get_env_extra().items():
        value = os.environ.get(ename)
        if value:
            res[key] = value
    return res


def set_extra_to_environ(key, value):
    os.environ[get_env_extra().get(key, key)] = value
